==> heptapod-rails-0.23.0.tar.gz: heptapod-rails-0.23.0/app/services/projects/update_remote_mirror_service.rb <==
# frozen_string_literal: true

module Projects
  class UpdateRemoteMirrorService < BaseService
    include Gitlab::Utils::StrongMemoize

    MAX_TRIES = 3

    def execute(remote_mirror, tries)
      return success unless remote_mirror.enabled?

      # Blocked URLs are a hard failure, no need to attempt to retry
      if Gitlab::UrlBlocker.blocked_url?(normalized_url(remote_mirror.url))
        hard_retry_or_fail(remote_mirror, _('The remote mirror URL is invalid.'), tries)
        return error(remote_mirror.last_error)
      end

      update_mirror(remote_mirror)

      success
    rescue Gitlab::Git::CommandError, Gitlab::Mercurial::HgGitRepository::HgError => e
      # This happens if one of the (h)gitaly or hg subprocess calls above fail,
      # for example when branches have diverged, or the pre-receive hook fails.
      hard_retry_or_fail(remote_mirror, e.message, tries)

      error(e.message)
    rescue StandardError => e
      remote_mirror.hard_fail!(e.message)
      raise e
    end

    private

    def normalized_url(url)
      strong_memoize(:normalized_url) do
        CGI.unescape(Gitlab::UrlSanitizer.sanitize(url))
      end
    end

    def update_mirror(remote_mirror)
      remote_mirror.update_start!
      remote_mirror.ensure_remote!

      # LFS objects must be sent first, or the push has dangling pointers
      send_lfs_objects!(remote_mirror) unless project.mercurial?

      branch_names = nil
      if project.mercurial? && remote_mirror.only_protected_branches
        # For now only pull protected branches to prevent pulling topics.
        # Later on, this will become a (non-default?) option.
        branch_names = project.repository.branches.map {|b| b.name}.select do |name|
          ProtectedBranch.protected?(project, name)
        end
      end

      response = remote_mirror.update_repository(gitlab_branches: branch_names)

      if response.divergent_refs.any?
        message = "Some refs have diverged and have not been updated on the remote:"
        message += "\n\n#{response.divergent_refs.join("\n")}"

        remote_mirror.mark_as_failed!(message)
      else
        remote_mirror.update_finish!
      end
    end

    def send_lfs_objects!(remote_mirror)
      return unless project.lfs_enabled?

      # TODO: Support LFS sync over SSH
      # https://gitlab.com/gitlab-org/gitlab/-/issues/249587
      return unless remote_mirror.url =~ /\Ahttps?:\/\//i
      return unless remote_mirror.password_auth?

      Lfs::PushService.new(
        project,
        current_user,
        url: remote_mirror.bare_url,
        credentials: remote_mirror.credentials
      ).execute
    end

    def hard_retry_or_fail(mirror, message, tries)
      if tries < MAX_TRIES
        mirror.hard_retry!(message)
      else
        # It's not likely we'll be able to recover from this ourselves, so we'll
        # notify the users of the problem, and don't trigger any sidekiq retries
        # Instead, we'll wait for the next change to try the push again, or until
        # a user manually retries.
        mirror.hard_fail!(message)
      end
    end
  end
end

==> heptapod-rails-0.24.0rc1.tar.gz: heptapod-rails-0.24.0rc1/app/services/projects/update_remote_mirror_service.rb <==
# frozen_string_literal: true

module Projects
  class UpdateRemoteMirrorService < BaseService
    include Gitlab::Utils::StrongMemoize

    MAX_TRIES = 3

    def execute(remote_mirror, tries)
      return success unless remote_mirror.enabled?

      # Blocked URLs are a hard failure, no need to attempt to retry
      if Gitlab::UrlBlocker.blocked_url?(normalized_url(remote_mirror.url))
        hard_retry_or_fail(remote_mirror, _('The remote mirror URL is invalid.'), tries)
        return error(remote_mirror.last_error)
      end

      update_mirror(remote_mirror)

      success
    rescue Gitlab::Git::CommandError, Gitlab::Mercurial::HgGitRepository::HgError => e
      # This happens if one of the (h)gitaly or hg subprocess calls above fail,
      # for example when branches have diverged, or the pre-receive hook fails.
      hard_retry_or_fail(remote_mirror, e.message, tries)

      error(e.message)
    rescue StandardError => e
      remote_mirror.hard_fail!(e.message)
      raise e
    end

    private

    def normalized_url(url)
      strong_memoize(:normalized_url) do
        CGI.unescape(Gitlab::UrlSanitizer.sanitize(url))
      end
    end

    def update_mirror(remote_mirror)
      remote_mirror.update_start!

      # LFS objects must be sent first, or the push has dangling pointers
      send_lfs_objects!(remote_mirror) unless project.mercurial?

      branch_names = nil
      if project.mercurial? && remote_mirror.only_protected_branches
        # For now only pull protected branches to prevent pulling topics.
        # Later on, this will become a (non-default?) option.
        branch_names = project.repository.branches.map {|b| b.name}.select do |name|
          ProtectedBranch.protected?(project, name)
        end
      end

      # Note: the Mercurial case doesn't care about inmemory_remote;
      #       the Git case doesn't care about gitlab_branches
      response = if Feature.enabled?(:update_remote_mirror_inmemory, project, default_enabled: :yaml)
                   remote_mirror.update_repository(inmemory_remote: true,
                                                   gitlab_branches: branch_names)
                 else
                   remote_mirror.ensure_remote!
                   remote_mirror.update_repository(inmemory_remote: false,
                                                   gitlab_branches: branch_names)
                 end

      if response.divergent_refs.any?
        message = "Some refs have diverged and have not been updated on the remote:"
        message += "\n\n#{response.divergent_refs.join("\n")}"

        remote_mirror.mark_as_failed!(message)
      else
        remote_mirror.update_finish!
      end
    end

    def send_lfs_objects!(remote_mirror)
      return unless project.lfs_enabled?

      # TODO: Support LFS sync over SSH
      # https://gitlab.com/gitlab-org/gitlab/-/issues/249587
      return unless remote_mirror.url =~ /\Ahttps?:\/\//i
      return unless remote_mirror.password_auth?

      Lfs::PushService.new(
        project,
        current_user,
        url: remote_mirror.bare_url,
        credentials: remote_mirror.credentials
      ).execute
    end

    def hard_retry_or_fail(mirror, message, tries)
      if tries < MAX_TRIES
        mirror.hard_retry!(message)
      else
        # It's not likely we'll be able to recover from this ourselves, so we'll
        # notify the users of the problem, and don't trigger any sidekiq retries
        # Instead, we'll wait for the next change to try the push again, or until
        # a user manually retries.
        mirror.hard_fail!(message)
      end
    end
  end
end
